from ._distribution import (
    AbstractDistribution as AbstractDistribution,
    AbstractMarginalDistribution as AbstractMarginalDistribution,
)
from ._gaussian_distributions import (
    IndependentFourierGaussian as IndependentFourierGaussian,
)
